// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

/**
 * Class containing a solver for hanjie puzzles. This solver uses methods that
 * only look at a single row or column (called a "line") at once. The solving
 * methods print messages on {@code System.out} if they find lines which are
 * impossible.
 */
public class HanjieSolver implements PuzzleSolver<Hanjie, HanjieState> {
    
    private Hanjie puzzle;
    private HanjieState puzzleState;
    //Properties of the puzzle
    private int height, width;
    private boolean hasColour;
    private int colourCount;
    //Solving info
    private int[][] state;
    private boolean solvingRows;
    private int currentLine;
    private boolean[] rowComplete;
    private boolean[] columnComplete;
    private int lastChange = -1;
    private boolean lastWasRow;
    
    /**
     * Constructs a new solver for the given puzzle.
     * 
     * @param puzzle the puzzle to solve
     */
    public HanjieSolver(Hanjie puzzle) {
        if (puzzle.getColourCount() > 32) throw new RuntimeException("Solver cannot handle puzzles with more than 32 colours.");
        this.puzzle = puzzle.isMutable() ? puzzle.makeImmutableCopy() : puzzle;
        puzzleState = new HanjieState(this.puzzle);
        height = this.puzzle.getHeight();
        width = this.puzzle.getWidth();
        hasColour = this.puzzle.hasColour();
        colourCount = this.puzzle.getColourCount();
        start();
    }
    
    /**
     * Constructs a new solver to work from a given state.
     * 
     * @param puzzleState the puzzle state to work from
     */
    public HanjieSolver(HanjieState puzzleState) {
        if (puzzleState.puzzle.getColourCount() > 32) throw new RuntimeException("Solver cannot handle puzzles with more than 32 colours.");
        this.puzzleState = puzzleState;
        puzzle = puzzleState.puzzle;
        height = puzzle.getHeight();
        width = puzzle.getWidth();
        hasColour = puzzle.hasColour();
        colourCount = puzzle.getColourCount();
        start();
    }
    
    public Hanjie getPuzzle() {
        return puzzle;
    }
    
    public HanjieState getState() {
        return puzzleState;
    }
    
    public void start() {
        solvingRows = true;
        currentLine = 0;
        rowComplete = new boolean[height];
        columnComplete = new boolean[width];
        state = new int[height][width];
        for (int i=0; i<height; i++) rowComplete[i] = false;
        for (int i=0; i<width; i++) columnComplete[i] = false;
        for (int i=0; i<height; i++) for (int j=0; j<width; j++)
        {
            state[i][j] = puzzleState.getState(i, j);
            if (state[i][j] < 0)
                state[i][j] = (1<<colourCount)-1;
            else
                state[i][j] = 1<<state[i][j];
        }
        lastChange = -1;
    }
    
    public boolean step() {
        boolean startSR = solvingRows;
        int startLine = currentLine;
        int temp;
        boolean hiddenChange = false;
        boolean changed = false;
        do {
            temp = solveLine();
            if (temp > 0) {
                startSR = solvingRows;
                startLine = currentLine;
            }
            changed = temp > 1;
            if (changed) {
                lastChange = currentLine;
                lastWasRow = solvingRows;
            }
            if (temp == -1) {
                if (solvingRows) {
                    System.out.println("Impossible line: Row " + (currentLine+1) + ".");
                } else {
                    System.out.println("Impossible line: Column " + (currentLine+1) + ".");
                }
            }
            nextLine();
        } while (!changed && !(solvingRows == startSR && currentLine == startLine));
        if (!changed) lastChange = -1;
        return changed;
    }
    
    /**
     * Performs a single pass through either the rows or columns.
     * 
     * @return {@code true} iff a change was made
     */
    public boolean pass() {
        boolean startSR = solvingRows;
        int temp;
        boolean changed = false;
        while (solvingRows == startSR) {
            temp = solveLine();
            changed |= temp > 1;
            if (temp > 1) {
                lastChange = currentLine;
                lastWasRow = solvingRows;
            } else if (temp == -1) {
                if (solvingRows) {
                    System.out.println("Impossible line: Row " + (currentLine+1) + ".");
                } else {
                    System.out.println("Impossible line: Column " + (currentLine+1) + ".");
                }
            }
            nextLine();
        }
        if (!changed) lastChange = -1;
        return changed;
    }
    
    public boolean solve() {
        boolean result = step();
        while (step());
        return result;
    }
    
    /**
     * Returns the location of the last change.
     * 
     * @return the row or column index of the last change or -1 if not
     *         applicable
     */
    public int getLastChange() {
        return lastChange;
    }
    
    /**
     * Returns whether or not the last change was a row.
     * 
     * @return {@code true} iff the last change was a row
     */
    public boolean getLastWasRow() {
        return lastWasRow;
    }
    
    /**
     * Advances to the next line (row/column).
     */
    public void nextLine() {
        currentLine++;
        if (solvingRows && currentLine == height) {
            currentLine = 0;
            solvingRows = false;
        } else if (!solvingRows && currentLine == width) {
            currentLine = 0;
            solvingRows = true;
        }
    }
    
    /**
     * Solves the current line (row/column) as much as possible.
     * 
     * @return {@code 0} if no change was made, {@code 1} if a hidden change
     *         was made, {@code 3} if a change was made or {@code -1} if the
     *         line was impossible.
     */
    public int solveLine() {
        int temp;
        int[] curLine;
        if (solvingRows) {
            if (rowComplete[currentLine]) {
                temp = 0;
            } else {
                curLine = new int[width];
                for (int i=0; i<width; i++) {
                    curLine[i] = state[currentLine][i];
                }
                temp = doLine(puzzle.getRowClue(currentLine), hasColour ? puzzle.getRowColour(currentLine) : null, curLine);
                rowComplete[currentLine] = true;
                for (int i=0; i<width; i++) {
                    state[currentLine][i] = curLine[i];
                    puzzleState.setState(currentLine, i, maskToColour(curLine[i]));
                    rowComplete[currentLine] &= (curLine[i] & (curLine[i]-1)) == 0;
                }
                rowComplete[currentLine] |= temp == -1;
            }
        } else {
            if (columnComplete[currentLine]) {
                temp = 0;
            } else {
                curLine = new int[height];
                for (int i=0; i<height; i++) {
                    curLine[i] = state[i][currentLine];
                }
                temp = doLine(puzzle.getColumnClue(currentLine), hasColour ? puzzle.getColumnColour(currentLine) : null, curLine);
                columnComplete[currentLine] = true;
                for (int i=0; i<height; i++) {
                    state[i][currentLine] = curLine[i];
                    puzzleState.setState(i, currentLine, maskToColour(curLine[i]));
                    columnComplete[currentLine] &= (curLine[i] & (curLine[i]-1)) == 0;
                }
                columnComplete[currentLine] |= temp == -1;
            }
        }
        return temp;
    }
    
    private int doLine(int[] clue, int[] clueColour, int[] line) {
        int changed = 0;
        int[] seen = new int[line.length];
        
        int oldLine;
        int[] leftMost = new int[clue.length];
        int i, j;
        for (int x=0; x<line.length; x++) {
            if ((line[x] & (line[x]-1)) == 0) {
                seen[x] = line[x];
                continue;
            }
            oldLine = line[x];
            for (int c=0; c<colourCount; c++) {
                if (canBeTest(seen[x], c) || !canBeTest(oldLine, c)) continue;
                line[x] = 1<<c;
                
                //We compute the leftmost possible position for each clue assuming line[x] has colour c
                //This is to determine whether such a position exists
                for (i=0; i<clue.length; i++) {
                    leftMost[i] = 0;
                }
                if (clue.length>0) {
                    for (j=clue[clue.length-1]; j<line.length; j++) if (!canBeTest(line[j], 0)) {
                        leftMost[clue.length-1] = j-clue[clue.length-1]+1;
                    }
                }
                i=0;
iloop:
                while (i<clue.length) {
                    //Must be after previous clue (with gap if same colour)
                    if (i>0) {
                        int fromPrev = leftMost[i-1] + clue[i-1] + (!hasColour || clueColour[i-1]==clueColour[i] ? 1 : 0);
                        if (fromPrev > leftMost[i]) leftMost[i] = fromPrev;
                    }
                    //Goes past end of line
                    if (leftMost[i]+clue[i] > line.length) break iloop;
                    //Pixels on this clue must be right colour
                    for (j=leftMost[i]; j<leftMost[i]+clue[i]; j++) {
                        if (!canBeTest(line[j], hasColour ? clueColour[i] : 1)) {
                            leftMost[i] = j+1;
                            if (leftMost[i]+clue[i] > line.length) break iloop;
                        }
                    }
                    //Previous clue must cover pixels before this which can't be 0
                    if (i>0) for (j=leftMost[i]-1; j>=leftMost[i-1]+clue[i-1]; j--) {
                        if (!canBeTest(line[j], 0)) {
                            leftMost[i-1] = j-clue[i-1]+1;
                            i--;
                            continue iloop;
                        }
                    }
                    //No previous clue so all pixels before must be 0
                    if (i==0) for (j=0; j<leftMost[i]; j++) if (!canBeTest(line[j], 0)) break iloop;
                    i++;
                }
                
                if (i==clue.length) { //It was possible
                    int fillMask = 1;
                    i = 0;
                    for (j=0; j<line.length; j++) {
                        if (i < clue.length && j == leftMost[i] + clue[i]) {
                            i++;
                            fillMask = 1;
                        }
                        if (i < clue.length && j == leftMost[i]) {
                            fillMask = 1 << (hasColour ? clueColour[i] : 1);
                        }
                        seen[j] |= fillMask;
                    }
                }
            }
            
            if (seen[x] == 0) {
                line[x] = oldLine;
                return -1;
            } else if ((seen[x] & (seen[x]-1)) == 0) {
                changed = 3;
            } else if (line[x] != seen[x]) {
                changed |= 1;
            }
            line[x] = seen[x];
        }
        return changed;
    }
    
    //Tests if an element of line (or equally state) indicates that it can be a specific colour.
    private static boolean canBeTest(int value, int colour) {
        return (value & 1<<colour) != 0;
    }
    
    private int maskToColour(int mask) {
        for (int j=0; j<colourCount; j++)
            if (mask == 1<<j) return j;
        return -1;
    }
    
}
