// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.twisty.
// 
// smithers.twisty is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.twisty is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.twisty. If not, see <http://www.gnu.org/licenses/>.

package smithers.twisty;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Shape;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Random;
import smithers.java3d.Drawable3D;
import smithers.java3d.Image3D;
import smithers.java3d.Panel3D;
import smithers.java3d.Polygon3D;
import smithers.java3d.Transformation;
import smithers.util.Launcher;

/**
 * Displays a {@code TwistyPuzzle} and allows the user to manipulate it. If the
 * puzzle is an {@code AnimatedTwistyPuzzle} then moves the user applies may be
 * animated over a specified length of time.
 */
public class TwistyPuzzleViewer extends Panel3D implements MouseListener, KeyListener, /*ActionListener,*/ Runnable {
    
    private static class StickeredFace extends Drawable3D {
        
        private float[][] outerPoly, innerPoly;
        
        private boolean usesImage = false;
        private BufferedImage image = null;
        private Image3D image3D = null;
        private Color fill = null;
        
        private StickeredFace(Polygon3D polygon, Color fill) {
            this.fill = fill;
            setPolygon(polygon);
        }
        
        private StickeredFace(Polygon3D polygon, BufferedImage image) {
            this.usesImage = true;
            this.image = image;
            setPolygon(polygon);
        }
        
        private void setPolygon(Polygon3D polygon) {
            poly = polygon;
            outerPoly = polygon.points();
            innerPoly = polygon.scale(1-grout).points();
            
            if (usesImage)
                image3D = new Image3D(image, innerPoly[0], innerPoly[1], innerPoly[innerPoly.length-1]);
        }
        
        public Shape drawOn(Graphics2D g, Transformation trans) {
            float[][] temp2D = trans.apply(outerPoly, null);
            Polygon outerPoly2D = new Polygon(new int[temp2D.length], new int[temp2D.length], 0);
            for (float[] p : temp2D)
                outerPoly2D.addPoint((int)p[0], (int)p[1]);
            g.setColor(Color.BLACK);
            g.fillPolygon(outerPoly2D);
            
            double area = 0; // Actually 2*area but I only need the sign
            for (int i=0; i<temp2D.length; i++)
                area += temp2D[i][0]*temp2D[(i+1)%temp2D.length][1] - temp2D[(i+1)%temp2D.length][0]*temp2D[i][1];
            
            if (area > 0) {
                if (usesImage) {
                    image3D.paint(g, trans);
                } else {
                    trans.apply(innerPoly, temp2D);
                    Polygon innerPoly2D = new Polygon(new int[temp2D.length], new int[temp2D.length], 0);
                    for (float[] p : temp2D)
                        innerPoly2D.addPoint((int)p[0], (int)p[1]);
                    g.setColor(fill);
                    g.fillPolygon(innerPoly2D);
                }
            }
            
            return outerPoly2D;
        }
        
    }
    
    private TwistyPuzzle puzzle;
    private static final float grout = 0.1f;
    private StickeredFace[] faces;
//     private TextField text;
    
    private boolean animated;
    private AnimatedTwistyPuzzle aniPuzzle; // equal to puzzle or null
    private Thread animationThread;
    private double aniTime;
    private Queue<Twist> twistQueue = new LinkedList<Twist>();
    
    private boolean demo;
    private RubiksCube rubiksPuzzle; // equal to puzzle or null
    private int s;
    
    /**
     * Constructs a new {@code TwistyPuzzleViewer} with the specified puzzle.
     * This viewer will display the given puzzle and allow the user to
     * manipulate it.
     * 
     * @param puzzle the puzzle to display and manipulate
     */
    public TwistyPuzzleViewer(TwistyPuzzle puzzle) {
        super((float)puzzle.getScale());
        this.puzzle = puzzle;
        this.faces = new StickeredFace[puzzle.stickerCount()];
        for (int i=0; i<puzzle.stickerCount(); i++) {
            faces[i] = new StickeredFace(puzzle.getPolygon(i), new Color(puzzle.getColor(i), true));
            faces[i].addMouseListener(this);
            add(faces[i]);
        }
        
        this.animated = false;
        this.demo = false;
        
        addKeyListener(this);
//         text = new TextField(40);
//         add(text);
//         text.addActionListener(this);
    }
    
    /**
     * Constructs a new {@code TwistyPuzzleViewer} with an animated puzzle.
     * This viewer will display the given puzzle and allow the user to
     * manipulate it. If {@code aniTime} is nonzero, then moves will be
     * animated over the given length of time.
     * 
     * @param puzzle  the puzzle to display and manipulate
     * @param aniTime the time to take to animate each move, if the puzzle can
     *                be animated
     */
    public TwistyPuzzleViewer(AnimatedTwistyPuzzle puzzle, double aniTime) {
        this(puzzle);
        
        this.aniPuzzle = puzzle;
        
        if (aniTime != 0.0) {
            this.animated = true;
            this.aniTime = aniTime;
        }
    }
    
    /**
     * Constructs a new {@code TwistyPuzzleViewer} with a Rubik's cube. This
     * viewer will display the given puzzle and allow the user to manipulate
     * it. If {@code aniTime} is nonzero, then moves will be animated over the
     * given length of time. If {@code demo} is {@code true}, the puzzle will
     * manipulate itself at random.
     * 
     * @param puzzle    the puzzle to display and manipulate
     * @param aniTime   the time to take to animate each move, if the puzzle
     *                  can be animated
     * @param imageName the image file to load sticker images from, or {@code
     *                  null}
     * @param demo      whether to demonstrate the puzzle moving on its own
     */
    public TwistyPuzzleViewer(RubiksCube puzzle, double aniTime, String imageName, boolean demo) {
        super((float)puzzle.getScale());
        this.puzzle = this.aniPuzzle = this.rubiksPuzzle = puzzle;
        this.s = puzzle.side();
        this.faces = new StickeredFace[puzzle.stickerCount()];
        
        BufferedImage[] stickerImages = null;
        if (imageName != null)
            stickerImages = loadStickerImages(imageName);
        if (stickerImages != null) {
            for (int i=0; i<puzzle.stickerCount(); i++) {
                faces[i] = new StickeredFace(puzzle.getPolygon(i), stickerImages[i]);
                faces[i].addMouseListener(this);
                add(faces[i]);
            }
        } else {
            for (int i=0; i<puzzle.stickerCount(); i++) {
                faces[i] = new StickeredFace(puzzle.getPolygon(i), new Color(puzzle.getColor(i), true));
                faces[i].addMouseListener(this);
                add(faces[i]);
            }
        }
        
        this.animated = aniTime != 0.0;
        this.aniTime = aniTime;
        this.demo = demo;
        
        addKeyListener(this);
    }
    
    private BufferedImage[] loadStickerImages(String fileName) {
        BufferedImage[] stickerImage = new BufferedImage[s*s*6];
        try {
            BufferedImage image = javax.imageio.ImageIO.read(new java.io.File(fileName));
            int w = image.getWidth();
            int h = image.getHeight();
            int tileSize = w/4 < h/3 ? w/4 : h/3;
            int sSize = tileSize / s;
            for (int i=0; i<s; i++) for (int j=0; j<s; j++) {
                stickerImage[s*s*0+s*j+i] = image.getSubimage(tileSize*1 + sSize*i, tileSize*0 + sSize*j, sSize, sSize);
                stickerImage[s*s*1+s*j+i] = image.getSubimage(tileSize*0 + sSize*i, tileSize*1 + sSize*j, sSize, sSize);
                stickerImage[s*s*2+s*j+i] = image.getSubimage(tileSize*1 + sSize*i, tileSize*1 + sSize*j, sSize, sSize);
                stickerImage[s*s*3+s*j+i] = image.getSubimage(tileSize*2 + sSize*i, tileSize*1 + sSize*j, sSize, sSize);
                stickerImage[s*s*4+s*j+i] = image.getSubimage(tileSize*3 + sSize*i, tileSize*1 + sSize*j, sSize, sSize);
                stickerImage[s*s*5+s*j+i] = image.getSubimage(tileSize*1 + sSize*i, tileSize*2 + sSize*j, sSize, sSize);
            }
            return stickerImage;
        } catch (Exception e) {
            return null;
        }
    }
    
    public void start() {
        animationThread = new Thread(this);
        animationThread.setDaemon(true);
        animationThread.start();
    }
    
    public void run() {
        if (! (animated || demo))
            return;
        
        Random r = null;
        double[] demoAngle = null;
        double t = 0.0;
        
        if (demo) {
            r = new Random();
            demoAngle = new double[2];
            demoAngle[0] = r.nextDouble() * 2 * Math.PI;
            demoAngle[1] = r.nextDouble() * 2 * Math.PI;
        }
        
        while (Thread.currentThread() == animationThread) {
            try {
                Thread.sleep(10);
            } catch (InterruptedException e) {
                break;
            }
            if (demo) {
                if (animated && !rubiksPuzzle.isAnimating() && twistQueue.isEmpty()) {
                    int face = r.nextInt(6);
                    int depth = r.nextInt(s);
                    twistQueue.add(rubiksPuzzle.getTwist(face, depth, depth+1, 1));
                }
                t += 0.002;
                if (t >= 1.0) {
                    t = 0.0;
                    demoAngle[0] = demoAngle[1];
                    demoAngle[1] = r.nextDouble() * 2 * Math.PI;
                    if (!animated) {
                        int face = r.nextInt(6);
                        int depth = r.nextInt(s);
                        rubiksPuzzle.twist(face, depth, depth+1, 1);
                    }
                }
                rotateView((float)(0.005 * ((1.0 - t) * Math.cos(demoAngle[0]) + t * Math.cos(demoAngle[1]))),
                           (float)(0.005 * ((1.0 - t) * Math.sin(demoAngle[0]) + t * Math.sin(demoAngle[1]))),
                           0);
            }
            if (animated) {
                if (!twistQueue.isEmpty() && !aniPuzzle.isAnimating()) {
                    aniPuzzle.twist(twistQueue.remove(), aniTime);
                }
                aniPuzzle.advanceAnimation(0.01);
            }
            repaint();
        }
    }
    
    public void repaint() {
        for (int i=0; i<puzzle.stickerCount(); i++) {
            faces[i].setPolygon(puzzle.getPolygon(i));
        }
        super.repaint();
    }
    
    public void mouseClicked(MouseEvent e) {
        Object source = e.getSource();
        if (!(source instanceof StickeredFace)) return;
        
        for (int i=0; i<faces.length; i++) if (faces[i] == source) {
            int dir = e.getButton();
            dir = dir==MouseEvent.BUTTON1 ? 0 : dir==MouseEvent.BUTTON3 ? 1 : 2;
            Twist twist = puzzle.getTwist(i, dir);
            if (animated) {
                twistQueue.add(twist);
            } else {
                puzzle.twist(twist);
                repaint();
            }
            return;
        }
    }
    
    public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_S) {
            puzzle.scramble();
            repaint();
        }
    }
    
//     public void actionPerformed(ActionEvent e) {
//         if(e.getSource().equals(text)) {
//             if(animated) {
//                 Twist[] twists = puzzle.getTwists(text.getText());
//                 for(Twist twist : twists) twistQueue.add(twist);
//             } else {
//                 puzzle.twist(text.getText());
//             }
//             text.setText("");
//             repaint();
//         }
//     }
    
    /* Specified by MouseListener */
    public void mousePressed(MouseEvent e) {}
    public void mouseReleased(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    
    /* Specified by KeyListener */
    public void keyTyped(KeyEvent e) {}
    public void keyReleased(KeyEvent e) {}
    
    public static void main(String[] args) {
        double aniTime = -1.0;
        String fName = null;
        boolean demo = false;
        TwistyPuzzle puzzle = new RubiksCube();
        String title = "Java Rubik's Cube by Smithers";
        
        for (int i=0; i<args.length; i++) {
            if (args[i].matches("-r.*")) {
                if (args[i].length()>2) puzzle = new RubiksCube(Integer.parseInt(args[i].substring(2)));
                else puzzle = new RubiksCube();
                title = "Java Rubik's Cube by Smithers";
            } else if (args[i].equals("-s")) {
                puzzle = new SquareOne();
                title = "Java Square One by Smithers";
            } else if (args[i].equals("--snake")) {
                puzzle = new Snake();
                title = "Java Rubik's Snake by Smithers";
            } else if (args[i].matches("-d") || args[i].matches("--demo")) {
                demo = true;
            } else if (args[i].matches("-a.*")) {
                aniTime = args[i].length()>2 ? Double.parseDouble(args[i].substring(2)) : 0.0;
            } else if (args[i].equals("-i")) {
                fName = args[++i];
            } else if (args[i].equals("-h") || args[i].equals("--help")) {
                System.out.println("smithers.twisty.TwistyPuzzleViewer -- Rubik's Cubes and similar puzzles.");
                System.out.println("Options:");
                System.out.println("  -r[s] Set puzzle to Rubik's Cube with side length s (default is 3).");
                System.out.println("  -s    Set puzzle to Square One / Cube21.");
                System.out.println("  --snake");
                System.out.println("        Set puzzle to Rubik's Snake.");
                System.out.println("  -a[t] Turn animation on (with time t) or off (-a) (does not apply to Rubik's Snake).");
                System.out.println("  -d");
                System.out.println("  --demo");
                System.out.println("        Run in demonstration mode (Rubik's Cube only).");
                System.out.println("  -i fName");
                System.out.println("        Get the face images from the image file fName (Rubik's Cube only).");
                System.out.println("  -h");
                System.out.println("  --help");
                System.out.println("        Display this help message and exit.");
                return;
            }
        }
        
        TwistyPuzzleViewer viewer;
        if (puzzle instanceof RubiksCube)
            viewer = new TwistyPuzzleViewer((RubiksCube)puzzle, aniTime, fName, demo);
        else if (puzzle instanceof AnimatedTwistyPuzzle)
            viewer = new TwistyPuzzleViewer((AnimatedTwistyPuzzle)puzzle, aniTime);
        else
            viewer = new TwistyPuzzleViewer(puzzle);
        Launcher.launch(viewer, title, 500, 500);
        if (viewer.animated || viewer.demo) viewer.start();
    }
    
}
